#coding:utf-8
"""Модуль nna содержит определения классов-оберток для объектов, используемых
в скриптах автоматизации ректификационной установки (вариант LITE): датчики,
цифровые модули, силовые модули и т.д.
Список изменений и история - в конце файла
:author: OldBean
21.10.2018"""
import sys
import smbus # В этом модуле - средства для работы с шиной i2c
import time # В этом модуле - задержки
import Adafruit_BMP.BMP085 as bmp # Обслуживание датчика давления
import math # В этом модуле - математика 
#-------------------------------------------------------------------------------
# Константы
ON = 100
OFF = 0
#-------------------------------------------------------------------------------
class PM(object):
  """Базовый класс-обертка исполнительного устройства (силового модуля). 
  Реализует базовые операции:
  1) проверка наличия и работоспособности устройства на шине i2c;
  2) отправка данных устройству при присваивании значения свойству v
     (количество байтов данных указывается параметром vlen
  3) включение устройства на максимум (метод on()) и выключение (метод off())"""
  def __new__(cls, addr, vlen = 1, bus = smbus.SMBus(1)):
    try: # Проверим наличие и работоспособность цифрового модуля
      bus.write_byte(addr, 0x00)
      # Все в порядке - создаем объект - цифровой модуль
      return super(PM, cls).__new__(cls)
    except:
      print("Силовой модуль модуль по адресу %x отсутствует или неисправен"%(addr))
      return None
  def __init__(self, addr, vlen = 1, bus = smbus.SMBus(1)):
    """Конструктор класса PM
    :param addr: адрес исполнительного устройства на шине i2c
    :param vlen: длина слова в байтах, передаваемых устройству (по умолчанию - 1)
    :param rmss: объект датчика RMS"""
    self._addr = addr
    self._bus = bus
    self._vlen = vlen
    self._v = 0 # Данные для обмена с устройством (vlen байтов)
  @property
  def v(self): 
    """ Возвращает буфер обмена (байт))"""
    return self._v
  @v.setter
  def v(self, value):
    """При присваивании значения, отправляет слово данных исполнительному
    устройству
    :param value: данные, отправляемые силовому модулю"""
    self._v = value
    if self._vlen == 1: # Данные - 1 байт
      self._bus.write_byte(self._addr, value & 0xff)
    else: # Данные - несколько байтов
      self._bus.read_byte(self._addr) # Сброс индекса байтов буфера устройства
      for i in range(self._vlen): # Передаем байты, начиная с младшего
        self._bus.write_byte(self._addr, value & 0xff)
        value >>= 8
  def on(self):
    """Включение исполнительного устройства на максимум"""
    self.v = 0xffffffff >> 8*(4 - self._vlen)
  def off(self):
    """Выключение выключение исполнительного устройства"""
    self.v = 0
#-------------------------------------------------------------------------------
class PWM(PM):
  """Класс-обертка для исполнительных устройств с ШИМ-регулированием (PWM)
  (шаг 0.1%)"""
  def __new__(cls, addr, qmax = 1, unit = '', bus = smbus.SMBus(1)):
    return super(PWM, cls).__new__(cls, addr, vlen = 2, bus = bus)
  def __init__(self, addr, qmax = 1, unit = '', bus = smbus.SMBus(1)):
    """Конструктор класса PWM
    :param addr: адрес исполнительного устройства на шине i2c
    :param qmax: максимальное значение физической величины, регулируемой
                 исполнительным устройством (при 100% ШИМ, код - 1000)
    :param rmss: объект датчика RMS"""
    self._coef = 1000.0/qmax # Для пересчета в физической величины в код
    self._unit = unit # Единица измерения
    self._q = 0.0 # Текущее значение физической величины
    super(PWM, self).__init__(addr, vlen = 2, bus = bus) # Конструктор базового класса
  @property
  def q(self):
    """Возвращает значение ранее установленной физической величины"""
    return self._q
  @q.setter
  def q(self, value):
    """Устанавливает значение физической величины и посылает исполнительному
    устройству код, соответствующий этой физической величине"""
    self._q = value
    self.v = int(value*self._coef + 0.5) # Код исполнительного устройства
  @property
  def u(self): 
    """Возвращает единицу физической величины, регулируемой данным исполнительным
    устройством"""
    return self._unit
#-------------------------------------------------------------------------------
class HTR(PM):
  """Класс-обертка для силовых модулей, регулирующих работу ТЭНов (HeaTeR).
  Пусть w - трбуемая мощность нагрева, w0 - номинальная мощность ТЭНа при
  220В, U - среднеквадратичное напряжение в сети на данный момент. Тогда код,
  посылаемый исполнительному стройству вычисляется по следующей формуле:
  pdm = 100*(w/w0)*(220^2/U^2)."""
  def __new__(cls, addr, w0 = 1500, rmss = None, bus = smbus.SMBus(1)):
    return super(HTR, cls).__new__(cls, addr, vlen = 1, bus = bus)
  def __init__(self, addr, w0 = 1500, rmss = None, bus = smbus.SMBus(1)):
    """Конструктор класса HTR
    :param addr: адрес исполнительного устройства на шине i2c
    :param w0: номинальная мощность ТЭНа
    :param rmss: объект датчика RMS"""
    self._addr = addr
    self._w0 = w0
    self._c0 = 100.0/w0 # Константа для вычисления pdm
    self._rmss = rmss
    if rmss is None:
      print('Отсутствует датчик RMS - стабилизации мощности нагрева не будет')
      self._c1 = 1.0
    else:
      self._c1 = 242000000.0/(rmss.coef*rmss.coef)
    self._rmss = rmss
    self._bus = bus
    self._w = 0 # Текущая мощность нагрева (Вт)
    super(HTR, self).__init__(addr, vlen = 1, bus = bus) # Конструктор базового класса
  @property
  def w(self):
    """Возвращает текущее значение мощности нагрева"""
    return self._w
  @w.setter
  def w(self, value):
    """Устанавливает значение физической величины и посылает исполнительному
    устройству код, соответствующий этой физической величине"""
    self._w = value
    pdm = value*self._c0
    if not self._rmss is None: # Стабилизация мощности нагрева
      crms = self._rmss.crms
      pdm = pdm*self._c1/crms
    self.v = int(pdm + 0.5) # Код исполнительного устройства
#-------------------------------------------------------------------------------
class Sensor(object):
  """Класс-обертка Sensor используется для всех датчикив, подключаемых к 
  цифровому модулю: термодатчики DS18B20, релейные датчики, любые аналоговые,
  датчики, снабженные 1-Wire адаптером, реализующим протокол прикладного уровня,
  используемый при работе с датчиками DS18B20."""
  def __init__(self, coef = 0, unit = None, offset = 0):
    """Конструктор класса.
    :param coef: коэффициент для перевода кода АЦП в физическое значение
    :param unit: единица измерения (текстовая строка)
    :param offset: смещение для коррекции нуля кода АЦП"""
    self._coef = coef
    self._unit = unit
    self._offset = offset
    self._code = 0x0000 # Текущее значение кода АЦП датчика
  @property
  def v(self): 
    """ Возвращает значение физической величины, измеренное датчиком."""
    return (self._code - self._offset)*self._coef
  @property
  def c(self): 
    """ Возвращает код АЦП датчика"""
    return self._code
  @c.setter
  def c(self, code):
    """ Устанавливает текущее значение кода АЦП датчика
    :param code: код АЦП, получаемый от цифрового модуля"""
    self._code = code
  @property
  def u(self):
    """ Возвращает единицу измерения (строку) физической величины, измеряемой
    датчиком"""
    return self._unit
  def __str__(self):
    """ Возвращает строку, содержащую всю информацию о датчике и текущем
    измеренной значении"""
    v = (self._code - self._offset)*self._coef
    if self._unit is None:
      return 'code: 0x%04x, value: %d'%(self._code, v) 
    return 'code: 0x%04x, value: % 3.3f, unit: %s'%(self._code, v, self._unit)
#-------------------------------------------------------------------------------
class DM(object):
  """Класс-обертка DM для цифрового модуля представляет собой контейнеры, 
  содержащие объекты-датчики, подключенные к цифровому модулю, текущие значения
  (коды АЦП) и общие параметры, такие, как адрес на шине i2c и т.д."""
  def __new__(cls, addr, bus = smbus.SMBus(1)):
    try: # Проверим наличие и работоспособность цифрового модуля
      bus.write_byte(addr, 0x00)
      # Все в порядке - создаем объект - цифровой модуль
      return super(DM, cls).__new__(cls) 
    except: 
      print("Цифровой модуль по адресу %x отсутствует или неисправен"%(addr))
      return None
  def __init__(self, addr, bus = smbus.SMBus(1)):
    """ addr - адрес цифрового модуля на шине i2c
       bus - объект-шина SMBus. По умолчанию - SMBus(1)."""
    self._addr = addr
    self._bus = bus
    self._buf = [0]*16 # Буфер данных, полученных от цифрового модуля
    self._sensors = [None]*8
  def update(self):
    """Чтение всего буфера данных цифрового модуля"""
    tries = 100 # Максимальное количество попыток доступа к данным модуля
    while tries > 0: 
      try: # Пытаемся получить данные...
        # Сбрасываем индекс байтов в буфере цифрового модуля
        self._bus.write_byte(self._addr, 0x00)
        for i in range(16): # Теперь считываем буфер байт за байтом
          self._buf[i] = self._bus.read_byte(self._addr)
        break
      except IOError:   # Устройство не готово - подождем 0.1 сек
        time.sleep(0.1) # и опять попробуем
        tries -= 1
    if tries <= 0: # Ну что-то серьезно не так - аварийное заверщение...
      print 'Цифровой модуль не доступен по i2c. Завершение работы.'
      sys.exit(1)
    for i in range(8): # Актуализируем данные подключенных датчиков
      if not self._sensors[i] is None:
        self._sensors[i].c = self._buf[2*i]*256 + self._buf[2*i + 1]
  def addSensor(self, ch = 0, sensor = None):
    """Добавление датчика к цифровому модулю
    :param ch: номер канала,цифрового модуля к которому подключен датчик
    :param sensor: датчик (объект класса Sensor)"""
    if self._sensors[ch] is None:
      self._sensors[ch] = sensor
    else:
      print('Канал цифрового модуля %d уже занят. Подключение отменено.')
  @property
  def sensors(self): 
    """ Возвращает список подключенных датчиков"""
    return self._sensors

  def __str__(self):
    """Сериализация всех данных, полученных от цифрового модуля"""
    s = 'Цифровой модуль:\n'
    for i in range(8):
      # Формируем слово из двух соседних байтов - это полный код АЦП
#      w = self._buf[2*i]*256 + self._buf[2*i + 1]
      s += '%d: '%(i)
      if self._sensors[i] is None: s += 'Нет датчика\n'
      else: s += str(self._sensors[i]) + '\n'
    return s
#-------------------------------------------------------------------------------
class BMP180(object):
  """Класс-обертка для датчика атмосферного давления и температуры BMP180.
     Давления и температура представлены соответствущими свойствами (P и T)
     экземпляра класса. Чтение датчиков происходит при каждом использовании
     свойств в правой части выражения."""
  sensor = None
  def __new__(cls):
    """ Перед созданием объекта В методе __new__ производится проверка
        наличия и работоспособности датчика BMP180"""
    try:
      cls.sensor = bmp.BMP085()
      cls.sensor.read_pressure()
      cls.sensor.read_temperature()
      return super(BMP180, cls).__new__(cls) 
    except: 
      print "Датчик BMP180 отсутствует или неисправен"
      return None
  def __init__(self):
    self._sensor = BMP180.sensor
  @property
  def p(self):
    """Свойство p - значение атмосферного давления в мм.рт.ст."""
    return self._sensor.read_pressure()/133.322 # Паскали -> в мм.рт.ст.
  @property
  def t(self):
    """Свойство t - значение температуры датчика в градусах Цельсия."""
    return self._sensor.read_temperature()
#-------------------------------------------------------------------------------
class RMS(object):
  """Класс-обертка для датчика среднеквадратичного напряжения в сети.
     Напряжение представлено двумя свойствами: crms - код rms - сумма 5000
     квадратов выборок напряжения и rms - собственно среднеквадратичное
     напряжение."""
  def __new__(cls, addr, coef = 0.3838, bus = smbus.SMBus(1)):
    try: # Проверим наличие и работоспособность цифрового модуля
      bus.write_byte(addr, 0x00)
      # Все в порядке - создаем объект - цифровой модуль
      return super(RMS, cls).__new__(cls) 
    except: 
      print("Датчик RMS по адресу %x отсутствует или неисправен"%(addr))
      return None
  def __init__(self, addr, coef = 0.3838, bus = smbus.SMBus(1)):
    """Конструктор класса RMS
    :param addr: адрес датчика RMS на шине i2c
    :param coef: калибровочный коэффициент для перевода кода в вольты
                 coef = Uобр/code, где Uобр - среднеквадратичное напряжение
                 сети, измеренное образцовым вольтметром TrueRMS, а code -
                 свойство code объекта-датчика RMS, полученного при измерении
                 этого же непряжения датчиком RMS
    :param bus: объект-шина SMBus. По умолчанию - SMBus(1)."""
    self._addr = addr
    self._coef = coef
    self._bus = bus
    self._crms = 0 # код rms - сумма 5000 квадратов выборок напряжения
    self._rms = 0 # собственно среднеквадратичное напряжение
  @property
  def crms(self):
    """Свойство crms - значение кода rms."""
    """Процедура получения данных с датчика RMS выглядит следующим образом:
       Сначала пишем в устройство нулевой байт. Тем самым мы устанавливаем
       индекс байтов буфера датчика RMS на младший байт. Затем последовательно
       читаем 4 байта (начиная с младшего) и формируем из него 32-разрядное
       слово - код RMS"""
    self._bus.write_byte(self._addr, 0x00)
    wrk = self._bus.read_byte(self._addr)
    wrk |= self._bus.read_byte(self._addr) << 8
    wrk |= self._bus.read_byte(self._addr) << 16
    wrk |= self._bus.read_byte(self._addr) << 24
    self._crms = wrk
    return self._crms
  @property
  def rms(self):
    """Свойство rms - значение среднеквадратичного напряжения"""
    return self._coef*math.sqrt(self.crms*0.0002)
  @property
  def coef(self):
    """Нормировочный коэффициент"""
    return self._coef
  @property
  def code(self):
    """Свойство code нужно для калибровки датчика RMS"""
    return math.sqrt(self.crms*0.0002)
#-------------------------------------------------------------------------------
"""На будущее
1. Подумать и реализовать блокировки по шине ic2, чтобы избежать конфликтов
   модулей на шине. Это может понадобиться при разработке приложений с GUI
"""
#-------------------------------------------------------------------------------
""" Список изменений и история
15.10.2018 - Написаны и отлажены классы-обертки Sensor и DM.
16.10.2018 - В модуль добавлен класс-обертка BMP180, косметические правки,
             в описание классов добавлены подробные комментарии.
19.10.2018 - 1. В конструктор класса Sensor добавлен параметр offset для
                компенсации нуля. По умолчанию offset = 0
             2. Реализован класс RMS для датчика среднеквадратичного напряжения
                питающей сети
21.10.2018   1. Реализован базовый класс силовых модулей - PM
             2. Реализован класс PWM для исполнительных устройств с ШИМ
                регулированием (с шагом 0.1%)
             3. Реализован класс HTR для контроллера ТЭНа
"""
